from typing import Any, Dict, List

from rlgym_learn.standard_impl.ppo import PPOMetricsLogger
from rlgym_learn.util import reporting


class ExampleLogger(PPOMetricsLogger[None]):

    def collect_state_metrics(self, data: List[None]) -> Dict[str, Any]:
        return {}

    def report_metrics(
        self,
        agent_controller_name,
        state_metrics,
        agent_metrics,
        wandb_run,
    ):
        report = {
            **agent_metrics,
            **state_metrics,
        }
        reporting.report_metrics(
            agent_controller_name, report, None, wandb_run=wandb_run
        )


def build_rlgym_v2_env():
    import numpy as np
    from rlgym.api import RLGym
    from rlgym.rocket_league import common_values
    from rlgym.rocket_league.action_parsers import LookupTableAction, RepeatAction
    from rlgym.rocket_league.done_conditions import (
        AnyCondition,
        GoalCondition,
        NoTouchTimeoutCondition,
        TimeoutCondition,
    )
    from rlgym.rocket_league.obs_builders import DefaultObs
    from rlgym.rocket_league.reward_functions import (
        CombinedReward,
        GoalReward,
        TouchReward,
    )
    from rlgym.rocket_league.sim import RocketSimEngine
    from rlgym.rocket_league.state_mutators import (
        FixedTeamSizeMutator,
        KickoffMutator,
        MutatorSequence,
    )

    spawn_opponents = True
    team_size = 2
    blue_team_size = team_size
    orange_team_size = team_size if spawn_opponents else 0
    action_repeat = 8
    no_touch_timeout_seconds = 30
    game_timeout_seconds = 300

    action_parser = RepeatAction(LookupTableAction(), repeats=action_repeat)
    termination_condition = GoalCondition()
    truncation_condition = AnyCondition(
        NoTouchTimeoutCondition(timeout_seconds=no_touch_timeout_seconds),
        TimeoutCondition(timeout_seconds=game_timeout_seconds),
    )

    reward_fn = CombinedReward((GoalReward(), 10), (TouchReward(), 0.1))

    obs_builder = DefaultObs(
        zero_padding=team_size,
        pos_coef=np.asarray(
            [
                1 / common_values.SIDE_WALL_X,
                1 / common_values.BACK_NET_Y,
                1 / common_values.CEILING_Z,
            ]
        ),
        ang_coef=1 / np.pi,
        lin_vel_coef=1 / common_values.CAR_MAX_SPEED,
        ang_vel_coef=1 / common_values.CAR_MAX_ANG_VEL,
        boost_coef=1 / 100.0,
    )

    state_mutator = MutatorSequence(
        FixedTeamSizeMutator(blue_size=blue_team_size, orange_size=orange_team_size),
        KickoffMutator(),
    )
    return RLGym(
        state_mutator=state_mutator,
        obs_builder=obs_builder,
        action_parser=action_parser,
        reward_fn=reward_fn,
        termination_cond=termination_condition,
        truncation_cond=truncation_condition,
        transition_engine=RocketSimEngine(),
    )


# The obs_space_type and action_space_type are determined by your choice of ObsBuilder and ActionParser respectively.
# The logic used here assumes you are using the types defined by the DefaultObs and LookupTableAction above.
def actor_factory(obs_space_type, action_space_type, device):
    from rlgym_learn.standard_impl.ppo import DiscreteFF

    obs_space_size = obs_space_type[1]
    action_space_size = action_space_type[1]
    return DiscreteFF(
        obs_space_size, action_space_size, [2048, 2048, 1024, 1024], device
    )


# The obs_space_type is determined by your choice of ObsBuilder.
# The logic used here assumes you are using the types defined by the DefaultObs above.
def critic_factory(obs_space_type, device):
    from rlgym_learn.standard_impl.ppo import BasicCritic

    obs_space_size = obs_space_type[1]
    return BasicCritic(obs_space_size, [2048, 2048, 1024, 1024], device)


if __name__ == "__main__":
    import numpy as np

    from rlgym_learn.learning_coordinator import LearningCoordinator
    from rlgym_learn.learning_coordinator_config import (
        BaseConfigModel,
        LearningCoordinatorConfigModel,
        ProcessConfigModel,
        generate_config,
        SerdeTypesModel,
        PyAnySerdeType
    )
    from rlgym_learn.standard_impl import (
        WandbMetricsLogger,
        WandbMetricsLoggerConfigModel,
    )
    from rlgym_learn.standard_impl.ppo import (
        ExperienceBufferConfigModel,
        GAETrajectoryProcessor,
        GAETrajectoryProcessorConfigModel,
        NumpyExperienceBuffer,
        PPOAgentController,
        PPOAgentControllerConfigModel,
        PPOLearnerConfigModel,
    )

    # Create the config that will be used for the run
    config = LearningCoordinatorConfigModel(
        base_config=BaseConfigModel(
            serde_types=SerdeTypesModel(
                agent_id_serde_type=PyAnySerdeType.STRING(),
                action_serde_type=PyAnySerdeType.NUMPY(np.int64),
                obs_serde_type=PyAnySerdeType.NUMPY(np.float64),
                reward_serde_type=PyAnySerdeType.FLOAT(),
                obs_space_serde_type=PyAnySerdeType.TUPLE(
                    (PyAnySerdeType.STRING(), PyAnySerdeType.INT())
                ),
                action_space_serde_type=PyAnySerdeType.TUPLE(
                    (PyAnySerdeType.STRING(), PyAnySerdeType.INT())
                ),
                state_metrics_serde_type=PyAnySerdeType.LIST(
                    PyAnySerdeType.NUMPY(np.float64)
                ),
            ),
            timestep_limit=1_000_000_000,  # Train for 1B steps
        ),
        process_config=ProcessConfigModel(
            n_proc=32,  # Number of processes to spawn to run environments. Increasing will use more RAM but should increase steps per second, up to a point
        ),
        agent_controllers_config={
            "PPO1": PPOAgentControllerConfigModel(
                log_to_wandb=True,  # logs run data to wandb
                learner_config=PPOLearnerConfigModel(
                    ent_coef=0.01,  # Sets the entropy coefficient used in the PPO algorithm
                    actor_lr=5e-5,  # Sets the learning rate of the actor model
                    critic_lr=5e-5,  # Sets the learning rate of the critic model
                ),
                experience_buffer_config=ExperienceBufferConfigModel(
                    max_size=150_000,  # Sets the number of timesteps to store in the experience buffer. Old timesteps will be pruned to only store the most recently obtained timesteps.
                    trajectory_processor_config=GAETrajectoryProcessorConfigModel(),
                ),
                wandb_config=WandbMetricsLoggerConfigModel(
                    group="rlgym-learn-testing", resume=True
                ),
            )
        },
        agent_controllers_save_folder="agent_controllers_checkpoints",  # (default value) WARNING: THIS PROCESS MAY DELETE ANYTHING INSIDE THIS FOLDER. This determines the parent folder for the runs for each agent controller. The runs folder for the agent controller will be this folder and then the agent controller config key as a subfolder.
    )

    # Generate the config file
    generate_config(
        learner_config=config,
        config_location="config.json",
        force_overwrite=True,
    )

    learning_coordinator = LearningCoordinator(
        build_rlgym_v2_env,
        agent_controllers={
            "PPO1": PPOAgentController(
                actor_factory=actor_factory,
                critic_factory=critic_factory,
                experience_buffer=NumpyExperienceBuffer(GAETrajectoryProcessor()),
                metrics_logger=WandbMetricsLogger(PPOMetricsLogger()),
                obs_standardizer=None,
            )
        },
        collect_state_metrics_fn=None,
        config_location="config.json",
    )
    learning_coordinator.start()



from dataclasses import dataclass
from typing import Generic, Optional

from rlgym.api import ActionType, AgentID, ObsType, RewardType


@dataclass
class Timestep(Generic[AgentID, ObsType, ActionType, RewardType]):
    __slots__ = (
        "env_id",
        "timestep_id",
        "previous_timestep_id",
        "agent_id",
        "obs",
        "next_obs",
        "action",
        "reward",
        "terminated",
        "truncated",
    )
    env_id: str
    timestep_id: int
    previous_timestep_id: Optional[int]
    agent_id: AgentID
    obs: ObsType
    next_obs: ObsType
    action: ActionType
    reward: RewardType
    terminated: bool
    truncated: bool



from __future__ import annotations

from abc import abstractmethod
from collections.abc import Callable
from datetime import timedelta
from multiprocessing import Process
from socket import _RetAddress, socket
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    Generic,
    List,
    Optional,
    Set,
    Tuple,
    Type,
    TypeVar,
    Union,
    _TypedDict,
)

from numpy import DTypeLike, _ShapeType, dtype, ndarray
from rlgym.api import (
    ActionSpaceType,
    ActionType,
    AgentID,
    EngineActionType,
    ObsSpaceType,
    ObsType,
    RewardType,
    RLGym,
    StateType,
)
from rlgym.rocket_league.api import Car, GameConfig, GameState, PhysicsObject

from rlgym_learn.api import ActionAssociatedLearningData, AgentController, StateMetrics
from rlgym_learn.experience import Timestep
from rlgym_learn.standard_impl import BatchRewardTypeNumpyConverter

if TYPE_CHECKING:
    from torch import Tensor

    from rlgym_learn.standard_impl.ppo import Trajectory

class EnvAction: ...

class EnvActionResponseType:
    STEP = ...
    RESET = ...
    SET_STATE = ...

class EnvActionResponse_STEP:
    def __new__(cls) -> EnvActionResponse_STEP: ...

class EnvActionResponse_RESET:
    def __new__(cls) -> EnvActionResponse_RESET: ...

class EnvActionResponse_SET_STATE(Generic[AgentID, StateType]):
    def __new__(
        cls,
        desired_state: StateType,
        prev_timestep_id_dict: Optional[Dict[AgentID, Optional[int]]],
    ) -> EnvActionResponse_SET_STATE[AgentID, StateType]: ...

class EnvActionResponse(Generic[AgentID, StateType]):
    STEP: Type[EnvActionResponse_STEP] = ...
    RESET: Type[EnvActionResponse_RESET] = ...
    SET_STATE: Type[EnvActionResponse_SET_STATE] = ...
    def enum_type(self) -> EnvActionResponseType: ...
    def desired_state(self) -> Optional[StateType]: ...
    def prev_timestep_id_dict(self) -> Optional[Dict[AgentID, Optional[int]]]: ...

class DerivedGAETrajectoryProcessorConfig:
    def __new__(
        cls, gamma: float, lmbda: float, dtype: dtype
    ) -> DerivedGAETrajectoryProcessorConfig: ...

class EnvProcessInterface(
    Generic[
        AgentID,
        ObsType,
        ActionType,
        EngineActionType,
        RewardType,
        StateType,
        ObsSpaceType,
        ActionSpaceType,
        StateMetrics,
        ActionAssociatedLearningData,
    ]
):
    def __new__(
        cls,
        agent_id_serde: PyAnySerdeType[AgentID],
        action_serde: PyAnySerdeType[ActionType],
        obs_serde: PyAnySerdeType[ObsType],
        reward_serde: PyAnySerdeType[RewardType],
        obs_space_serde: PyAnySerdeType[ObsSpaceType],
        action_space_serde: PyAnySerdeType[ActionSpaceType],
        state_serde_option: Optional[PyAnySerdeType[StateType]],
        state_metrics_serde_option: Optional[PyAnySerdeType[StateMetrics]],
        recalculate_agent_id_every_step: bool,
        flinks_folder_option: str,
        min_process_steps_per_inference: int,
        send_state_to_agent_controllers: bool,
        should_collect_state_metrics: bool,
    ) -> EnvProcessInterface: ...
    def init_processes(
        self, proc_package_defs: List[Process, socket, _RetAddress, str]
    ) -> Tuple[
        Dict[str, Tuple[List[AgentID], List[ObsType]]],
        Dict[str, Tuple[Optional[StateType], None, None]],
        ObsSpaceType,
        ActionSpaceType,
    ]: ...
    def add_process(
        self, proc_package_def: Tuple[Process, socket, _RetAddress, str]
    ): ...
    def delete_process(self): ...
    def increase_min_process_steps_per_inference(self) -> int: ...
    def decrease_min_process_steps_per_inference(self) -> int: ...
    def cleanup(self): ...
    def collect_step_data(
        self,
    ) -> Tuple[
        int,
        Dict[str, Tuple[List[AgentID], List[ObsType]]],
        Dict[
            str,
            Tuple[
                List[Timestep],
                ActionAssociatedLearningData,
                Optional[StateMetrics],
                Optional[StateType],
            ],
        ],
        Dict[
            str,
            Tuple[
                Optional[StateType],
                Optional[Dict[AgentID, bool]],
                Optional[Dict[AgentID, bool]],
            ],
        ],
    ]: ...
    def send_env_actions(self, env_actions: Dict[str, EnvAction]): ...

class AgentManager(
    Generic[
        AgentID,
        ObsType,
        ActionType,
        RewardType,
        StateType,
        ObsSpaceType,
        ActionSpaceType,
        StateMetrics,
        ActionAssociatedLearningData,
    ]
):
    def __new__(
        cls,
        agent_controllers: List[
            AgentController[
                Any,
                AgentID,
                ObsType,
                ActionType,
                RewardType,
                StateType,
                ObsSpaceType,
                ActionSpaceType,
                StateMetrics,
                ActionAssociatedLearningData,
                Any,
            ],
        ],
        batched_tensor_action_associated_learning_data: bool,
    ) -> AgentManager: ...
    def get_env_actions(
        self, env_obs_data_dict: Dict[str, Tuple[List[AgentID], List[ObsType]]]
    ) -> Dict[str, EnvAction]: ...

class GAETrajectoryProcessor:
    def __new__(
        cls, batch_reward_type_numpy_converter: BatchRewardTypeNumpyConverter
    ) -> GAETrajectoryProcessor: ...
    def load(self, config: DerivedGAETrajectoryProcessorConfig): ...
    def process_trajectories(
        self, trajectories: List[Trajectory], return_std: ndarray
    ) -> Tuple[
        List[AgentID],
        List[ObsType],
        List[ActionType],
        Tensor,
        Tensor,
        ndarray,
        ndarray,
        ndarray,
    ]: ...

def env_process(
    proc_id: str,
    child_end,
    parent_sockname,
    build_env_fn: Callable[
        [],
        RLGym[
            AgentID,
            ObsType,
            ActionType,
            EngineActionType,
            RewardType,
            StateType,
            ObsSpaceType,
            ActionSpaceType,
        ],
    ],
    flinks_folder: str,
    shm_buffer_size: int,
    agent_id_serde: PyAnySerdeType[AgentID],
    action_serde: PyAnySerdeType[ActionType],
    obs_serde: PyAnySerdeType[ObsType],
    reward_serde: PyAnySerdeType[RewardType],
    obs_space_serde: PyAnySerdeType[ObsSpaceType],
    action_space_serde: PyAnySerdeType[ActionSpaceType],
    state_serde_option: Optional[PyAnySerdeType[StateType]],
    state_metrics_serde_option: Optional[PyAnySerdeType[StateMetrics]],
    collect_state_metrics_fn_option: Optional[
        Callable[[StateType, Dict[AgentID, RewardType]], StateMetrics]
    ],
    send_state_to_agent_controllers: bool,
    render: bool,
    render_delay_option: Optional[timedelta],
    recalculate_agent_id_every_step: bool,
): ...
def recvfrom_byte(socket: socket): ...
def sendto_byte(socket: socket, address: _RetAddress): ...

T = TypeVar("T")
KeysT = TypeVar("KeysT")
ValuesT = TypeVar("ValuesT")

class PythonSerde(Generic[T]):
    @abstractmethod
    def to_bytes(self, obj: T) -> bytes:
        """
        Function to convert obj to bytes, for passing between batched agent and the agent manager.
        :return: bytes b such that from_bytes(b) == obj.
        """
        raise NotImplementedError

    @abstractmethod
    def from_bytes(self, byts: bytes) -> T:
        """
        Function to convert bytes to T, for passing between batched agent and the agent manager.
        :return: T obj such that from_bytes(to_bytes(obj)) == obj.
        """
        raise NotImplementedError

class PickleableInitStrategy(Generic[T]):
    def __new__(cls, init_strategy: InitStrategy[T]) -> PickleableInitStrategy[T]: ...

class InitStrategy(Generic[T]):
    ALL: Type[InitStrategy_ALL] = ...
    SOME: Type[InitStrategy_SOME] = ...
    NONE: Type[InitStrategy_NONE] = ...

class InitStrategy_ALL(InitStrategy[T]):
    def __new__(cls) -> InitStrategy_ALL: ...

class InitStrategy_SOME(InitStrategy[T]):
    def __new__(cls, kwargs: List[str]) -> InitStrategy_ALL:
        """
        kwargs: a list of keyword arguments to pass to the constructor of the dataclass
        """
        ...

class InitStrategy_NONE(InitStrategy[T]):
    def __new__(cls) -> InitStrategy_NONE: ...

class PickleablePyAnySerdeType(Generic[T]):
    def __new__(
        cls, pyany_serde_type: PyAnySerdeType[T]
    ) -> PickleablePyAnySerdeType[T]: ...

class PyAnySerdeType(Generic[T]):
    BOOL: Type[PyAnySerdeType_BOOL] = ...
    BYTES: Type[PyAnySerdeType_BYTES] = ...
    COMPLEX: Type[PyAnySerdeType_COMPLEX] = ...
    DATACLASS: Type[PyAnySerdeType_DATACLASS] = ...
    DICT: Type[PyAnySerdeType_DICT] = ...
    DYNAMIC: Type[PyAnySerdeType_DYNAMIC] = ...
    FLOAT: Type[PyAnySerdeType_FLOAT] = ...
    INT: Type[PyAnySerdeType_INT] = ...
    LIST: Type[PyAnySerdeType_LIST] = ...
    NUMPY: Type[PyAnySerdeType_NUMPY] = ...
    OPTION: Type[PyAnySerdeType_OPTION] = ...
    PICKLE: Type[PyAnySerdeType_PICKLE] = ...
    PYTHONSERDE: Type[PyAnySerdeType_PYTHONSERDE] = ...
    SET: Type[PyAnySerdeType_SET] = ...
    STRING: Type[PyAnySerdeType_STRING] = ...
    TUPLE: Type[PyAnySerdeType_TUPLE] = ...
    TYPEDDICT: Type[PyAnySerdeType_TYPEDDICT] = ...
    UNION: Type[PyAnySerdeType_UNION] = ...

class PyAnySerdeType_BOOL(PyAnySerdeType[bool]):
    def __new__(cls) -> PyAnySerdeType_BOOL: ...

class PyAnySerdeType_BYTES(PyAnySerdeType[bytes]):
    def __new__(cls) -> PyAnySerdeType_BYTES: ...

class PyAnySerdeType_COMPLEX(PyAnySerdeType[complex]):
    def __new__(cls) -> PyAnySerdeType_COMPLEX: ...

class PyAnySerdeType_DATACLASS(PyAnySerdeType[T]):
    def __new__(
        cls,
        clazz: T,
        init_strategy: InitStrategy,
        field_serde_type_dict: Dict[str, PyAnySerdeType],
    ) -> PyAnySerdeType_DATACLASS[T]:
        """
        clazz: the dataclass to be serialized
        init_strategy: defines the initialization strategy
        field_serde_type_dict: dict to define the serde to be used with each field in the dataclass
        """
        ...

class PyAnySerdeType_DICT(PyAnySerdeType[Dict[KeysT, ValuesT]]):
    def __new__(
        cls,
        keys_serde_type: PyAnySerdeType[KeysT],
        values_serde_type: PyAnySerdeType[ValuesT],
    ) -> PyAnySerdeType_DICT[KeysT, ValuesT]: ...

class PyAnySerdeType_DYNAMIC(PyAnySerdeType[Any]):
    def __new__(cls) -> PyAnySerdeType_DYNAMIC: ...

class PyAnySerdeType_FLOAT(PyAnySerdeType[float]):
    def __new__(cls) -> PyAnySerdeType_FLOAT: ...

class PyAnySerdeType_INT(PyAnySerdeType[int]):
    def __new__(cls) -> PyAnySerdeType_INT: ...

class PyAnySerdeType_LIST(PyAnySerdeType[List[T]]):
    def __new__(cls, items_serde_type: PyAnySerdeType[T]) -> PyAnySerdeType_LIST[T]: ...

class PyAnySerdeType_NUMPY(PyAnySerdeType[ndarray[_ShapeType, DTypeLike]]):
    def __new__(
        cls, dtype: DTypeLike
    ) -> PyAnySerdeType_NUMPY[_ShapeType, DTypeLike]: ...

class PyAnySerdeType_OPTION(PyAnySerdeType[Optional[T]]):
    def __new__(
        cls, value_serde_type: PyAnySerdeType[T]
    ) -> PyAnySerdeType_OPTION[T]: ...

class PyAnySerdeType_PICKLE(PyAnySerdeType[Any]):
    def __new__(cls) -> PyAnySerdeType_PICKLE: ...

class PyAnySerdeType_PYTHONSERDE(PyAnySerdeType[T]):
    def __new__(
        cls, python_serde_type: PythonSerde[T]
    ) -> PyAnySerdeType_PYTHONSERDE[T]: ...

class PyAnySerdeType_SET(PyAnySerdeType[Set[T]]):
    def __new__(cls, items_serde_type: PyAnySerdeType[T]) -> PyAnySerdeType_SET[T]: ...

class PyAnySerdeType_STRING(PyAnySerdeType[str]):
    def __new__(cls) -> PyAnySerdeType_STRING: ...

class PyAnySerdeType_TUPLE(PyAnySerdeType[Tuple]):
    def __new__(item_serde_types: Tuple[PyAnySerdeType]) -> PyAnySerdeType_TUPLE: ...

class PyAnySerdeType_TYPEDDICT(PyAnySerdeType[_TypedDict]):
    def __new__(
        key_serde_type_dict: Dict[str, PyAnySerdeType]
    ) -> PyAnySerdeType_TYPEDDICT: ...

class PyAnySerdeType_UNION(PyAnySerdeType[Union]):
    def __new__(
        option_serde_types: List[PyAnySerdeType], option_choice_fn: Callable[[Any], int]
    ) -> PyAnySerdeType_UNION: ...



from .agent_controller import AgentController, DerivedAgentControllerConfig
from .typing import (
    ActionAssociatedLearningData,
    AgentControllerConfig,
    AgentControllerData,
    StateMetrics,
)



from __future__ import annotations

from typing import Optional

import numpy as np
from rlgym.api import AgentID
from rlgym.rocket_league.api import Car, GameConfig, GameState, PhysicsObject

from rlgym_learn import InitStrategy, PyAnySerdeType

# class RocketLeaguePyAnySerdeType_GAME_CONFIG(PyAnySerdeType[GameConfig]):
#     def __init__(self) -> RocketLeaguePyAnySerdeType_GAME_CONFIG:
#         return PyAnySerdeType.DATACLASS(
#             GameConfig,
#             InitStrategy.NONE,
#             {
#                 "gravity": PyAnySerdeType.FLOAT(),
#                 "boost_consumption": PyAnySerdeType.FLOAT(),
#                 "dodge_deadzone": PyAnySerdeType.FLOAT(),
#             },
#         )


def game_config_serde() -> PyAnySerdeType[GameConfig]:
    return PyAnySerdeType.DATACLASS(
        GameConfig,
        InitStrategy.NONE(),
        {
            "gravity": PyAnySerdeType.FLOAT(),
            "boost_consumption": PyAnySerdeType.FLOAT(),
            "dodge_deadzone": PyAnySerdeType.FLOAT(),
        },
    )


# class RocketLeaguePyAnySerdeType_PHYSICS_OBJECT(PyAnySerdeType[PhysicsObject]):
#     def __init__(self) -> RocketLeaguePyAnySerdeType_PHYSICS_OBJECT:
#         return PyAnySerdeType.DATACLASS(
#             PhysicsObject,
#             InitStrategy.NONE,
#             {
#                 "position": PyAnySerdeType.NUMPY(np.float32),
#                 "linear_velocity": PyAnySerdeType.NUMPY(np.float32),
#                 "angular_velocity": PyAnySerdeType.NUMPY(np.float32),
#                 "_quaternion": PyAnySerdeType.OPTION(PyAnySerdeType.NUMPY(np.float32)),
#                 "_rotation_mtx": PyAnySerdeType.OPTION(
#                     PyAnySerdeType.NUMPY(np.float32)
#                 ),
#                 "_euler_angles": PyAnySerdeType.OPTION(
#                     PyAnySerdeType.NUMPY(np.float32)
#                 ),
#             },
#         )


def physics_object_serde() -> PyAnySerdeType[PhysicsObject]:
    return PyAnySerdeType.DATACLASS(
        PhysicsObject,
        InitStrategy.NONE(),
        {
            "position": PyAnySerdeType.NUMPY(np.float32),
            "linear_velocity": PyAnySerdeType.NUMPY(np.float32),
            "angular_velocity": PyAnySerdeType.NUMPY(np.float32),
            "_quaternion": PyAnySerdeType.OPTION(PyAnySerdeType.NUMPY(np.float32)),
            "_rotation_mtx": PyAnySerdeType.OPTION(PyAnySerdeType.NUMPY(np.float32)),
            "_euler_angles": PyAnySerdeType.OPTION(PyAnySerdeType.NUMPY(np.float32)),
        },
    )


# class RocketLeaguePyAnySerdeType_CAR(PyAnySerdeType[Car[AgentID]]):
#     def __init__(
#         self, agent_id_serde_type: PyAnySerdeType[AgentID]
#     ) -> RocketLeaguePyAnySerdeType_CAR[AgentID]:
#         return PyAnySerdeType.DATACLASS(
#             Car,
#             InitStrategy.NONE,
#             {
#                 "team_num": PyAnySerdeType.INT(),
#                 "hitbox_type": PyAnySerdeType.INT(),
#                 "ball_touches": PyAnySerdeType.INT(),
#                 "team_num": PyAnySerdeType.INT(),
#                 "bump_victim_id": PyAnySerdeType.OPTION(agent_id_serde_type),
#                 "demo_respawn_timer": PyAnySerdeType.FLOAT(),
#                 "wheels_with_contact": PyAnySerdeType.TUPLE(
#                     (
#                         PyAnySerdeType.BOOL(),
#                         PyAnySerdeType.BOOL(),
#                         PyAnySerdeType.BOOL(),
#                         PyAnySerdeType.BOOL(),
#                     )
#                 ),
#                 "supersonic_time": PyAnySerdeType.FLOAT(),
#                 "boost_amount": PyAnySerdeType.FLOAT(),
#                 "boost_active_time": PyAnySerdeType.FLOAT(),
#                 "handbrake": PyAnySerdeType.FLOAT(),
#                 "is_jumping": PyAnySerdeType.BOOL(),
#                 "has_jumped": PyAnySerdeType.BOOL(),
#                 "is_holding_jump": PyAnySerdeType.BOOL(),
#                 "jump_time": PyAnySerdeType.FLOAT(),
#                 "has_flipped": PyAnySerdeType.BOOL(),
#                 "has_double_jumped": PyAnySerdeType.BOOL(),
#                 "air_time_since_jump": PyAnySerdeType.FLOAT(),
#                 "flip_time": PyAnySerdeType.FLOAT(),
#                 "flip_torque": PyAnySerdeType.NUMPY(np.float32),
#                 "is_autoflipping": PyAnySerdeType.BOOL(),
#                 "autoflip_timer": PyAnySerdeType.FLOAT(),
#                 "autoflip_direction": PyAnySerdeType.FLOAT(),
#                 "physics": RocketLeaguePyAnySerdeType.PHYSICS_OBJECT(),
#                 "_inverted_physics": RocketLeaguePyAnySerdeType.PHYSICS_OBJECT(),
#             },
#         )


def car_serde(
    agent_id_serde_type: PyAnySerdeType[AgentID],
) -> PyAnySerdeType[Car[AgentID]]:
    return PyAnySerdeType.DATACLASS(
        Car,
        InitStrategy.NONE(),
        {
            "team_num": PyAnySerdeType.INT(),
            "hitbox_type": PyAnySerdeType.INT(),
            "ball_touches": PyAnySerdeType.INT(),
            "team_num": PyAnySerdeType.INT(),
            "bump_victim_id": PyAnySerdeType.OPTION(agent_id_serde_type),
            "demo_respawn_timer": PyAnySerdeType.FLOAT(),
            "wheels_with_contact": PyAnySerdeType.TUPLE(
                (
                    PyAnySerdeType.BOOL(),
                    PyAnySerdeType.BOOL(),
                    PyAnySerdeType.BOOL(),
                    PyAnySerdeType.BOOL(),
                )
            ),
            "supersonic_time": PyAnySerdeType.FLOAT(),
            "boost_amount": PyAnySerdeType.FLOAT(),
            "boost_active_time": PyAnySerdeType.FLOAT(),
            "handbrake": PyAnySerdeType.FLOAT(),
            "is_jumping": PyAnySerdeType.BOOL(),
            "has_jumped": PyAnySerdeType.BOOL(),
            "is_holding_jump": PyAnySerdeType.BOOL(),
            "jump_time": PyAnySerdeType.FLOAT(),
            "has_flipped": PyAnySerdeType.BOOL(),
            "has_double_jumped": PyAnySerdeType.BOOL(),
            "air_time_since_jump": PyAnySerdeType.FLOAT(),
            "flip_time": PyAnySerdeType.FLOAT(),
            "flip_torque": PyAnySerdeType.NUMPY(np.float32),
            "is_autoflipping": PyAnySerdeType.BOOL(),
            "autoflip_timer": PyAnySerdeType.FLOAT(),
            "autoflip_direction": PyAnySerdeType.FLOAT(),
            "physics": physics_object_serde(),
            "_inverted_physics": physics_object_serde(),
        },
    )


# class RocketLeaguePyAnySerdeType_GAME_STATE(PyAnySerdeType[GameState]):
#     def __init__(
#         self, agent_id_serde_type: PyAnySerdeType[AgentID]
#     ) -> RocketLeaguePyAnySerdeType_GAME_STATE:
#         return PyAnySerdeType.DATACLASS(
#             GameState,
#             InitStrategy.NONE,
#             {
#                 "tick_count": PyAnySerdeType.INT(),
#                 "goal_scored": PyAnySerdeType.BOOL(),
#                 "config": RocketLeaguePyAnySerdeType.GAME_CONFIG(),
#                 "cars": PyAnySerdeType.DICT(
#                     agent_id_serde_type,
#                     RocketLeaguePyAnySerdeType.CAR(agent_id_serde_type),
#                 ),
#                 "ball": RocketLeaguePyAnySerdeType.PHYSICS_OBJECT(),
#                 "_inverted_ball": RocketLeaguePyAnySerdeType.PHYSICS_OBJECT(),
#                 "boost_pad_timers": PyAnySerdeType.NUMPY(np.float32),
#                 "_inverted_boost_pad_timers": PyAnySerdeType.NUMPY(np.float32),
#             },
#         )


def game_state_serde(
    agent_id_serde_type: PyAnySerdeType[AgentID],
) -> PyAnySerdeType[GameState]:
    return PyAnySerdeType.DATACLASS(
        GameState,
        InitStrategy.NONE(),
        {
            "tick_count": PyAnySerdeType.INT(),
            "goal_scored": PyAnySerdeType.BOOL(),
            "config": game_config_serde(),
            "cars": PyAnySerdeType.DICT(
                agent_id_serde_type,
                car_serde(agent_id_serde_type),
            ),
            "ball": physics_object_serde(),
            "_inverted_ball": physics_object_serde(),
            "boost_pad_timers": PyAnySerdeType.NUMPY(np.float32),
            "_inverted_boost_pad_timers": PyAnySerdeType.NUMPY(np.float32),
        },
    )


# class RocketLeaguePyAnySerdeType:
#     GAME_CONFIG = RocketLeaguePyAnySerdeType_GAME_CONFIG
#     PHYSICS_OBJECT = RocketLeaguePyAnySerdeType_PHYSICS_OBJECT
#     CAR = RocketLeaguePyAnySerdeType_CAR
#     GAME_STATE = RocketLeaguePyAnySerdeType_GAME_STATE



from __future__ import annotations

import cProfile
import os
from collections.abc import Callable
from typing import Any, Dict, Generic, Optional, Union

from rlgym.api import (
    ActionSpaceType,
    ActionType,
    AgentID,
    EngineActionType,
    ObsSpaceType,
    ObsType,
    RewardType,
    RLGym,
    StateType,
)

from .agent import AgentManager
from .api import ActionAssociatedLearningData, AgentController, StateMetrics
from .env_processing import EnvProcessInterface
from .learning_coordinator_config import (
    DEFAULT_CONFIG_FILENAME,
    LearningCoordinatorConfigModel,
)
from .rlgym_learn import PyAnySerdeType
from .util import KBHit


class LearningCoordinator(
    Generic[
        AgentID,
        ObsType,
        ActionType,
        EngineActionType,
        RewardType,
        StateType,
        ObsSpaceType,
        ActionSpaceType,
        StateMetrics,
        ActionAssociatedLearningData,
    ]
):
    def __init__(
        self,
        env_create_function: Callable[
            [],
            RLGym[
                AgentID,
                ObsType,
                ActionType,
                EngineActionType,
                RewardType,
                StateType,
                ObsSpaceType,
                ActionSpaceType,
            ],
        ],
        agent_controllers: Dict[
            str,
            AgentController[
                Any,
                AgentID,
                ObsType,
                ActionType,
                RewardType,
                StateType,
                ObsSpaceType,
                ActionSpaceType,
                StateMetrics,
                ActionAssociatedLearningData,
                Any,
            ],
        ],
        collect_state_metrics_fn: Optional[
            Callable[[StateType, Dict[str, Any]], StateMetrics]
        ] = None,
        config_location: str = None,
    ):
        if config_location is None:
            config_location = os.path.join(os.getcwd(), DEFAULT_CONFIG_FILENAME)
        assert os.path.isfile(
            config_location
        ), f"{config_location} is not a valid location from which to read config, aborting."

        with open(config_location, "rt") as f:
            self.config = LearningCoordinatorConfigModel.model_validate_json(f.read())

        self.agent_manager = AgentManager(
            agent_controllers,
            self.config.base_config.batched_tensor_action_associated_learning_data,
        )

        self.cumulative_timesteps = 0
        self.env_process_interface = EnvProcessInterface(
            env_create_function,
            collect_state_metrics_fn,
            self.config.base_config.serde_types,
            self.config.process_config.min_process_steps_per_inference,
            self.config.base_config.send_state_to_agent_controllers,
            self.config.base_config.flinks_folder,
            self.config.base_config.shm_buffer_size,
            self.config.base_config.random_seed,
            self.config.process_config.recalculate_agent_id_every_step,
        )
        (
            initial_env_obs_data_dict,
            initial_state_info,
            obs_space,
            action_space,
        ) = self.env_process_interface.init_processes(
            n_processes=self.config.process_config.n_proc,
            spawn_delay=self.config.process_config.instance_launch_delay,
            render=self.config.process_config.render,
            render_delay=self.config.process_config.render_delay,
        )
        print("Loading agent controllers...")
        self.agent_manager.set_space_types(obs_space, action_space)
        self.agent_manager.load_agent_controllers(self.config)
        # Handle actions for observations created on process init
        self.initial_env_actions = self.agent_manager.get_env_actions(
            initial_env_obs_data_dict, initial_state_info
        )
        print("Learner successfully initialized!")
        # TODO: delete and remove import
        self.prof = cProfile.Profile()
        self.prof.enable()

    def start(self):
        """
        Function to wrap the _run function in a try/catch/finally
        block to ensure safe execution and error handling.
        :return: None
        """
        try:
            self._run()
        except Exception:
            import traceback

            print("\n\nLEARNING LOOP ENCOUNTERED AN ERROR\n")
            traceback.print_exc()

            try:
                self.save()
            except:
                print("FAILED TO SAVE ON EXIT")
                traceback.print_exc()

        finally:
            self.prof.disable()
            self.prof.dump_stats("ppo_prof.prof")
            self.cleanup()

    def _run(self):
        """
        Learning function. This is where the magic happens.
        :return: None
        """

        # Class to watch for keyboard hits
        kb = KBHit()
        print(
            "Press (p) to pause, (c) to checkpoint, (q) to checkpoint and quit (after next iteration)\n"
            + "(a) to add an env process, (d) to delete an env process\n"
            + "(j) to increase min inference size, (l) to decrease min inference size\n"
        )
        # Handle actions for observations created on process init
        self.env_process_interface.send_env_actions(self.initial_env_actions)

        # Collect the desired number of timesteps from our environments.
        loop_iterations = 0
        while self.cumulative_timesteps < self.config.base_config.timestep_limit:
            total_timesteps_collected, env_obs_data_dict, timestep_data, state_info = (
                self.env_process_interface.collect_step_data()
            )
            self.cumulative_timesteps += total_timesteps_collected
            self.agent_manager.process_timestep_data(timestep_data)

            self.env_process_interface.send_env_actions(
                self.agent_manager.get_env_actions(env_obs_data_dict, state_info)
            )
            loop_iterations += 1
            if loop_iterations % 50 == 0:
                if self.process_kbhit(kb):
                    break
        if self.cumulative_timesteps >= self.config.base_config.timestep_limit:
            print("Hit timestep limit, cleaning up...")
        else:
            print("Quitting and cleaning up...")

    def process_kbhit(self, kb: KBHit) -> bool:
        # Check if keyboard press
        # p: pause, any key to resume
        # c: checkpoint
        # q: checkpoint and quit

        if kb.kbhit():
            c = kb.getch()
            if c == "p":  # pause
                print("Paused, press any key to resume")
                while True:
                    if kb.kbhit():
                        break
            if c in ("c", "q"):
                self.agent_manager.save_agent_controllers()
            if c == "q":
                return True
            if c in ("c", "p"):
                print("Resuming...\n")
            if c == "a":
                print("Adding process...")
                self.env_process_interface.add_process()
                print(f"Process added. ({self.env_process_interface.n_procs} total)")
            if c == "d":
                print("Deleting process...")
                self.env_process_interface.delete_process()
                print(f"Process deleted. ({self.env_process_interface.n_procs} total)")
            if c == "j":
                min_process_steps_per_inference = (
                    self.env_process_interface.increase_min_process_steps_per_inference()
                )
                print(
                    f"Min process steps per inference increased to {min_process_steps_per_inference} ({(100 * min_process_steps_per_inference / self.env_process_interface.n_procs):.2f}% of processes)"
                )
            if c == "l":
                min_process_steps_per_inference = (
                    self.env_process_interface.decrease_min_process_steps_per_inference()
                )
                print(
                    f"Min process steps per inference decreased to {min_process_steps_per_inference} ({(100 * min_process_steps_per_inference / self.env_process_interface.n_procs):.2f}% of processes)"
                )
            return False

    def save(self):
        self.agent_manager.save_agent_controllers()

    def cleanup(self):
        """
        Function to clean everything up before shutting down.
        :return: None.
        """
        self.env_process_interface.cleanup()
        self.agent_manager.cleanup()



from .learning_coordinator import LearningCoordinator
from .learning_coordinator_config import (
    BaseConfigModel,
    LearningCoordinatorConfigModel,
    ProcessConfigModel,
    SerdeTypesModel,
    generate_config,
)
from .rlgym_learn import AgentManager as RustAgentManager
from .rlgym_learn import (
    DerivedGAETrajectoryProcessorConfig as RustDerivedGAETrajectoryProcessorConfig,
)
from .rlgym_learn import EnvAction, EnvActionResponse, EnvActionResponseType
from .rlgym_learn import EnvProcessInterface as RustEnvProcessInterface
from .rlgym_learn import GAETrajectoryProcessor as RustGAETrajectoryProcessor
from .rlgym_learn import (
    InitStrategy,
    PickleableInitStrategy,
    PickleablePyAnySerdeType,
    PyAnySerdeType,
)
from .rlgym_learn import env_process as rust_env_process
from .rlgym_learn import recvfrom_byte, sendto_byte



import json
import os
import time
from abc import abstractmethod
from dataclasses import dataclass
from typing import Any, Dict, Generic, List, Optional, TypeVar

from pydantic import BaseModel, Field

import wandb

from ..api import AgentControllerData, StateMetrics
from .dict_metrics_logger import DictMetricsLogger
from .metrics_logger import DerivedMetricsLoggerConfig, MetricsLogger

InnerMetricsLoggerConfig = TypeVar("InnerMetricsLoggerConfig")
InnerMetricsLoggerAdditionalDerivedConfig = TypeVar(
    "InnerMetricsLoggerAdditionalDerivedConfig"
)


def convert_nested_dict(d):
    new = {}
    for k, v in d.items():
        if isinstance(v, dict):
            converted = convert_nested_dict(v)
            to_add = {f"{k}/{k1}": v1 for k1, v1 in converted.items()}
        else:
            to_add = {k: v}
        new = {**new, **to_add}
    return new


class WandbMetricsLoggerConfigModel(BaseModel):
    enable: bool = True
    project: str = "rlgym-learn"
    group: str = "unnamed-runs"
    run: str = "rlgym-learn-run"
    id: Optional[str] = None
    new_run_with_timestamp_suffix: bool = False
    additional_wandb_run_config: Dict[str, Any] = Field(default_factory=dict)


@dataclass
class WandbAdditionalDerivedConfig(
    Generic[InnerMetricsLoggerConfig, InnerMetricsLoggerAdditionalDerivedConfig]
):
    derived_wandb_run_config: Dict[str, Any] = Field(default_factory=dict)
    timestamp_suffix: Optional[str] = None
    inner_metrics_logger_config: InnerMetricsLoggerConfig = None
    inner_metrics_logger_additional_derived_config: (
        InnerMetricsLoggerAdditionalDerivedConfig
    ) = None


class WandbMetricsLogger(
    MetricsLogger[
        WandbMetricsLoggerConfigModel,
        WandbAdditionalDerivedConfig[
            InnerMetricsLoggerConfig, InnerMetricsLoggerAdditionalDerivedConfig
        ],
        StateMetrics,
        AgentControllerData,
    ],
    Generic[
        InnerMetricsLoggerConfig,
        InnerMetricsLoggerAdditionalDerivedConfig,
        StateMetrics,
        AgentControllerData,
    ],
):
    def __init__(
        self,
        inner_metrics_logger: DictMetricsLogger[
            InnerMetricsLoggerConfig,
            InnerMetricsLoggerAdditionalDerivedConfig,
            StateMetrics,
            AgentControllerData,
        ],
        checkpoint_file_name: str = "wandb_metrics_logger.json",
    ):
        self.inner_metrics_logger = inner_metrics_logger
        self.checkpoint_file_name = checkpoint_file_name

    def collect_state_metrics(self, data: List[StateMetrics]):
        self.inner_metrics_logger.collect_state_metrics(data)

    def collect_agent_metrics(self, data: AgentControllerData):
        self.inner_metrics_logger.collect_agent_metrics(data)

    def report_metrics(self):
        self.wandb_run.log(convert_nested_dict(self.inner_metrics_logger.get_metrics()))
        self.inner_metrics_logger.report_metrics()

    def validate_config(self, config_obj: Any):
        return WandbMetricsLoggerConfigModel.model_validate(config_obj)

    def load(self, config):
        self.config = config
        self.inner_metrics_logger.load(
            DerivedMetricsLoggerConfig(
                checkpoint_load_folder=config.checkpoint_load_folder,
                agent_controller_name=config.agent_controller_name,
                metrics_logger_config=config.additional_derived_config.inner_metrics_logger_config,
                additional_derived_config=config.additional_derived_config.inner_metrics_logger_additional_derived_config,
            )
        )
        if self.config.checkpoint_load_folder is not None:
            self._load_from_checkpoint()
        if not self.config.metrics_logger_config.enable:
            self.wandb_run = None
            self.run_id = None
            return

        if (
            self.config.checkpoint_load_folder is not None
            and self.config.metrics_logger_config.id is not None
        ):
            if self.run_id is not None:
                print(
                    f"{self.config.agent_controller_name}: Wandb run id from checkpoint ({self.run_id}) is being overridden by wandb run id from config: {self.config.metrics_logger_config.id}"
                )
            self.run_id = self.config.metrics_logger_config.id
        else:
            self.run_id = None

        wandb_config = {
            **self.config.additional_derived_config.derived_wandb_run_config,
            **self.config.metrics_logger_config.additional_wandb_run_config,
        }

        run_name = self.config.metrics_logger_config.run
        if self.config.metrics_logger_config.new_run_with_timestamp_suffix:
            print(
                f"{self.config.agent_controller_name}: Due to config, a new wandb run is being created with timestamp suffix. This run will use the project and group specified in config, and will use the run name in config prepended to the timestamp suffix."
            )
            if (
                self.config.additional_derived_config.timestamp_suffix is not None
                and len(self.config.additional_derived_config.timestamp_suffix) > 0
            ):
                run_name += self.config.additional_derived_config.timestamp_suffix
            else:
                run_name += f"-{time.time_ns()}"

        self.wandb_run = wandb.init(
            project=self.config.metrics_logger_config.project,
            group=self.config.metrics_logger_config.group,
            config=wandb_config,
            name=run_name,
            id=self.run_id,
            resume="allow",
            reinit=True,
        )
        self.run_id = self.wandb_run.id
        print(f"{self.config.agent_controller_name}: Created wandb run! {self.run_id}")

    def _load_from_checkpoint(self):
        with open(
            os.path.join(
                self.config.checkpoint_load_folder,
                self.checkpoint_file_name,
            ),
            "rt",
        ) as f:
            state = json.load(f)
        if "run_id" in state:
            self.run_id = state["run_id"]
        else:
            self.run_id = None

    def save_checkpoint(self, folder_path):
        os.makedirs(folder_path, exist_ok=True)
        state = {"run_id": self.run_id}
        with open(
            os.path.join(
                folder_path,
                self.checkpoint_file_name,
            ),
            "wt",
        ) as f:
            json.dump(state, f, indent=4)



"""
File: reporting.py
Author: Matthew Allen

Description:
    Misc. functions to log a dictionary of metrics to wandb and print them to the console.
"""

import locale

import numpy as np
import torch

locale.setlocale(locale.LC_ALL, "")


def _form_printable_groups(report):
    """
    Function to create a list of dictionaries containing the data to print to the console in a specific order.
    :param report: A dictionary containing all of the values to organize.
    :return: A list of dictionaries containing keys organized in the desired fashion.
    """

    groups = [
        {
            "Average Undiscounted Episodic Return": report[
                "Average Undiscounted Episodic Return"
            ],
            "Actor Entropy": report["Actor Entropy"],
            "Critic Loss": report["Critic Loss"],
        },
        {
            "Mean KL Divergence": report["Mean KL Divergence"],
            "SB3 Clip Fraction": report["SB3 Clip Fraction"],
            "Actor Update Magnitude": report["Actor Update Magnitude"],
            "Critic Update Magnitude": report["Critic Update Magnitude"],
        },
        {
            "Collected Steps per Second": report["Collected Steps per Second"],
            "Overall Steps per Second": report["Overall Steps per Second"],
        },
        {
            "Timestep Collection Time": report["Timestep Collection Time"],
            "Timestep Consumption Time": report["Timestep Consumption Time"],
            "PPO Batch Consumption Time": report["PPO Batch Consumption Time"],
            "Total Iteration Time": report["Total Iteration Time"],
        },
        {
            "Cumulative Model Updates": report["Cumulative Model Updates"],
            "Cumulative Timesteps": report["Cumulative Timesteps"],
        },
        {"Timesteps Collected": report["Timesteps Collected"]},
    ]

    return groups


def report_metrics(
    agent_controller_name, loggable_metrics, debug_metrics, wandb_run=None
):
    """
    Function to report a dictionary of metrics to the console and wandb.
    :param loggable_metrics: Dictionary containing all the data to be logged.
    :param debug_metrics: Optional dictionary containing extra data to be printed to the console for debugging.
    :param wandb_run: Wandb run to log to.
    :return: None.
    """

    if wandb_run is not None:
        wandb_run.log(loggable_metrics)

    # Print debug data first.
    if debug_metrics is not None:
        print("\nBEGIN DEBUG\n")
        print(dump_dict_to_debug_string(debug_metrics))
        print("\nEND DEBUG\n")

    # Print the loggable metrics in a desirable format to the console.
    print(
        "{}{}{}".format(
            "-" * 8, f"BEGIN {agent_controller_name} ITERATION REPORT", "-" * 8
        )
    )
    groups = _form_printable_groups(loggable_metrics)
    out = ""
    for group in groups:
        out += dump_dict_to_debug_string(group) + "\n"
    print(out[:-2])
    print(
        "{}{}{}\n\n".format(
            "-" * 8, f"END {agent_controller_name} ITERATION REPORT", "-" * 8
        )
    )


def dump_dict_to_debug_string(dictionary):
    """
    Function to format the data in a loggable dictionary so the line length is limited.

    :param dictionary: Data to format.
    :return: A string containing the formatted elements of that dictionary.
    """

    debug_string = ""
    for key, val in dictionary.items():
        if type(val) == torch.Tensor:
            if len(val.shape) == 0:
                val = val.detach().cpu().item()
            else:
                val = val.detach().cpu().tolist()

        # Format lists of numbers as [num_1, num_2, num_3] where num_n is clipped at 5 decimal places.
        if type(val) in (tuple, list, np.ndarray):
            arr_str = []
            if type(val) != np.ndarray or val.shape:
                for arg in val:
                    arr_str.append(
                        locale.format_string("%7.5f", arg, grouping=True)
                        if type(arg) == float
                        else "{},".format(arg)
                    )
            else:
                arg = val.item()
                arr_str.append(
                    locale.format_string("%7.5f", arg, grouping=True)
                    if type(arg) == float
                    else "{},".format(arg)
                )

            arr_str = " ".join(arr_str)
            debug_string = "{}{}: [{}]\n".format(debug_string, key, arr_str[:-1])

        # Format floats such that only 5 decimal places are shown.
        elif type(val) in (float, np.float32, np.float64):

            debug_string = "{}{}: {}\n".format(
                debug_string, key, locale.format_string("%7.5f", val, grouping=True)
            )

        # Print ints with comma separated thousands (locale aware).
        elif type(val) in (int, np.int32, np.int64):
            debug_string = "{}{}: {}\n".format(
                debug_string, key, locale.format_string("%d", val, grouping=True)
            )
        # Default to just printing the value if it isn't a type we know how to format.
        else:
            debug_string = "{}{}: {}\n".format(debug_string, key, val)

    return debug_string



from __future__ import annotations

import random
import socket
from collections.abc import Callable
from dataclasses import dataclass
from datetime import timedelta
from typing import Dict, Optional

import numpy as np
from rlgym.api import (
    ActionSpaceType,
    ActionType,
    AgentID,
    EngineActionType,
    ObsSpaceType,
    ObsType,
    RewardType,
    RLGym,
    StateType,
)

from ..api import StateMetrics
from ..rlgym_learn import PickleablePyAnySerdeType
from ..rlgym_learn import env_process as rust_env_process
from ..rlgym_learn import recvfrom_byte, sendto_byte


@dataclass
class PickleableSerdeTypeConfig:
    agent_id_serde_type: PickleablePyAnySerdeType
    action_serde_type: PickleablePyAnySerdeType
    obs_serde_type: PickleablePyAnySerdeType
    reward_serde_type: PickleablePyAnySerdeType
    obs_space_serde_type: PickleablePyAnySerdeType
    action_space_serde_type: PickleablePyAnySerdeType
    state_serde_type: PickleablePyAnySerdeType
    state_metrics_serde_type: PickleablePyAnySerdeType


def env_process(
    proc_id: str,
    parent_sockname,
    build_env_fn: Callable[
        [],
        RLGym[
            AgentID,
            ObsType,
            ActionType,
            EngineActionType,
            RewardType,
            StateType,
            ObsSpaceType,
            ActionSpaceType,
        ],
    ],
    serde_type_config: PickleableSerdeTypeConfig,
    collect_state_metrics_fn: Optional[
        Callable[[StateType, Dict[AgentID, RewardType]], StateMetrics]
    ],
    send_state_to_agent_controllers: bool,
    flinks_folder: str,
    shm_buffer_size: int,
    seed: int,
    render_this_proc: bool,
    render_delay: Optional[float],
    recalculate_agent_id_every_step: bool,
):
    child_end = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    child_end.bind(("127.0.0.1", 0))

    random.seed(seed)
    np.random.seed(seed)

    sendto_byte(child_end, parent_sockname)
    recvfrom_byte(child_end)

    rust_env_process(
        proc_id,
        child_end,
        parent_sockname,
        build_env_fn,
        flinks_folder,
        shm_buffer_size,
        serde_type_config.agent_id_serde_type,
        serde_type_config.action_serde_type,
        serde_type_config.obs_serde_type,
        serde_type_config.reward_serde_type,
        serde_type_config.obs_space_serde_type,
        serde_type_config.action_space_serde_type,
        serde_type_config.state_serde_type,
        serde_type_config.state_metrics_serde_type,
        collect_state_metrics_fn,
        send_state_to_agent_controllers,
        render_this_proc,
        timedelta(seconds=render_delay),
        recalculate_agent_id_every_step,
    )



from typing import TypeVar

AgentControllerConfig = TypeVar("AgentControllerConfig")
AgentControllerData = TypeVar("AgentControllerData")
StateMetrics = TypeVar("StateMetrics")
ActionAssociatedLearningData = TypeVar("ActionAssociatedLearningData")
